# SKIP LICENSE INSERTION

"""
elytra-ms
~~~~~~~~~~~~~~~~~~~

A Python Library for various Microsoft APIs, including the Xbox and Bedrock Realms APIs.

(c) 2023-2024 AstreaTSS under MIT, see LICENSE for more details.
"""

from .bedrock_realms import *
from .const import *
from .core import *
from .protocols import *
from .xbox import *

__version__ = "0.3.2"
